/*
 * Copyright 1995, 1999 by Paul Mattes.
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted,
 *  provided that the above copyright notice appear in all copies and that
 *  both that copyright notice and this permission notice appear in
 *  supporting documentation.
 *
 * x3270, c3270, s3270 and tcl3270 are distributed in the hope that they will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the file LICENSE
 * for more details.
 */

/*
 *	objects.h
 *		x3270 object names.
 */

#define ObjConfirmButton	"confirmButton"
#define ObjConfirm2Button	"confirm2Button"
#define ObjCancelButton		"cancelButton"
#define ObjDialog		"dialog"
#define ObjSmallLabel		"smallLabel"
#define ObjNameLabel		"nameLabel"
#define ObjDataLabel		"dataLabel"
