/*
 * Copyright 1995, 1999, 2001 by Paul Mattes.
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted,
 *  provided that the above copyright notice appear in all copies and that
 *  both that copyright notice and this permission notice appear in
 *  supporting documentation.
 *
 * x3270, c3270, s3270, tcl3270 and pr3287 are distributed in the hope that
 * they will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * file LICENSE for more details.
 */

/*
 *	tablesc.h
 *		Global declarations for tables.c.
 */

extern unsigned char asc2cg[256];
extern unsigned char cg2asc[256];
extern unsigned char ebc2cg[256];
extern unsigned char ebc2cg0[256];
extern unsigned char cg2ebc[256];
extern unsigned char cg2ebc0[256];
extern unsigned char ebc2asc[256];
extern unsigned char ebc2asc0[256];
extern unsigned char ebc2asc7[256];
extern unsigned char ebc2asc70[256];
extern unsigned char asc2ebc[256];
extern unsigned char asc2ebc0[256];
extern unsigned char cg2uc[256];
extern unsigned char ge2asc[256];
extern unsigned char ge2cg8[256];
extern unsigned char asc2uc[256];
extern unsigned char ebc2uc[256];
extern unsigned char ft2asc[256];
extern unsigned char ft2asc0[256];
extern unsigned char asc2ft[256];
extern unsigned char asc2ft0[256];
